import logging
import sqlalchemy
from contextlib import contextmanager
from recap.analyzers.abstract import AbstractAnalyzer, BaseMetadataModel
from recap.browsers.db import create_browser, TablePath, ViewPath
from typing import Generator


log = logging.getLogger(__name__)


class Column(BaseMetadataModel):
    autoincrement: bool | None
    default: str | None
    nullable: bool
    type: str
    generic_type: str | None
    comment: str | None


class Columns(BaseMetadataModel):
    __root__: dict[str, Column] = {}


class TableColumnAnalyzer(AbstractAnalyzer):
    def __init__(self, engine: sqlalchemy.engine.Engine):
        self.engine = engine

    def analyze(
        self,
        path: TablePath | ViewPath,
    ) -> Columns | None:
        table = path.table if isinstance(path, TablePath) else path.view
        results = {}
        columns = sqlalchemy.inspect(self.engine).get_columns(
            table,
            path.schema_,
        )
        for column in columns:
            try:
                generic_type = column['type'].as_generic()
                # Strip length/precision to make generic strings more generic.
                if isinstance(generic_type, sqlalchemy.sql.sqltypes.String):
                    generic_type.length = None
                elif isinstance(generic_type, sqlalchemy.sql.sqltypes.Numeric):
                    generic_type.precision = None
                    generic_type.scale = None
                column['generic_type'] = str(generic_type)
            except NotImplementedError as e:
                # Unable to convert. Probably a weird type like PG's OID.
                log.debug(
                    'Unable to get generic type for path=%s column=%s',
                    path,
                    column.get('name', column),
                    exc_info=e,
                )
            # The `type` field is not JSON encodable; convert to string.
            column['type'] = str(column['type'])
            column_name = column['name']
            del column['name']
            results[column_name] = Column(
                autoincrement=column.get('autoincrement'),
                default=column['default'],
                generic_type=column.get('generic_type'),
                nullable=column['nullable'],
                type=str(column['type']),
                comment=column.get('comment')
            )
        if results:
            return Columns.parse_obj(results)
        return None


@contextmanager
def create_analyzer(**config) -> Generator['TableColumnAnalyzer', None, None]:
    with create_browser(**config) as browser:
        yield TableColumnAnalyzer(browser.engine)



from .abstract import AbstractCatalog
from contextlib import contextmanager
from recap.plugins import load_catalog_plugins
from typing import Generator


@contextmanager
def create_catalog(
    plugin: str = 'db',
    **config,
) -> Generator['AbstractCatalog', None, None]:
    catalog_plugins = load_catalog_plugins()
    catalog_plugin_module = catalog_plugins.get(plugin)
    assert catalog_plugin_module, \
        f"Unable to find catalog plugin module={catalog_plugin_module}"
    with catalog_plugin_module.create_catalog(**config) as catalog:
        yield catalog



import logging
import sqlalchemy
from contextlib import contextmanager
from recap.analyzers.abstract import AbstractAnalyzer, BaseMetadataModel
from recap.browsers.db import create_browser, TablePath, ViewPath
from typing import Any, Generator


log = logging.getLogger(__name__)


class UserAccess(BaseMetadataModel):
    privileges: list[str]
    read: bool
    write: bool


class Access(BaseMetadataModel):
    __root__: dict[str, UserAccess] = {}


class TableAccessAnalyzer(AbstractAnalyzer):
    def __init__(self, engine: sqlalchemy.engine.Engine):
        self.engine = engine

    def analyze(
        self,
        path: TablePath | ViewPath,
    ) -> Access | None:
        table = path.table if isinstance(path, TablePath) else path.view
        with self.engine.connect() as conn:
            results = {}
            try:
                rows = conn.execute(
                    "SELECT * FROM information_schema.role_table_grants "
                    "WHERE table_schema = %s AND table_name = %s",
                    path.schema_,
                    table,
                )
                for row in rows.all():
                    privilege_type = row['privilege_type']
                    user_grants: dict[str, Any] = results.get(row['grantee'], {
                        'privileges': [],
                        'read': False,
                        'write': False,
                    })
                    user_grants['privileges'].append(privilege_type)
                    if privilege_type == 'SELECT':
                        user_grants['read'] = True
                    if privilege_type in ['INSERT', 'UPDATE', 'DELETE', 'TRUNCATE']:
                        user_grants['write'] = True
                    results[row['grantee']] = user_grants
            except Exception as e:
                # TODO probably need a more tightly bound exception here
                # We probably don't have access to the information_schema, so
                # skip it.
                log.debug(
                    'Unable to fetch access for path=%s',
                    path,
                    exc_info=e,
                )
            if results:
                return Access.parse_obj(results)
            return None


@contextmanager
def create_analyzer(**config) -> Generator['TableAccessAnalyzer', None, None]:
    with create_browser(**config) as browser:
        yield TableAccessAnalyzer(browser.engine)



import logging
import sqlalchemy
from contextlib import contextmanager
from recap.analyzers.abstract import AbstractAnalyzer, BaseMetadataModel
from recap.browsers.db import create_browser, TablePath, ViewPath
from typing import Generator


log = logging.getLogger(__name__)


class Index(BaseMetadataModel):
    columns: list[str]
    unique: bool


class Indexes(BaseMetadataModel):
    __root__: dict[str, Index] = {}


class TableIndexAnalyzer(AbstractAnalyzer):
    def __init__(self, engine: sqlalchemy.engine.Engine):
        self.engine = engine

    def analyze(
        self,
        path: TablePath | ViewPath,
    ) -> Indexes | None:
        table = path.table if isinstance(path, TablePath) else path.view
        indexes = {}
        index_dicts = sqlalchemy.inspect(self.engine).get_indexes(
            table,
            path.schema_,
        )
        for index_dict in index_dicts:
            indexes[index_dict['name']] = {
                'columns': index_dict.get('column_names', []),
                'unique': index_dict['unique'],
            }
        if indexes:
            return Indexes.parse_obj(indexes)
        return None


@contextmanager
def create_analyzer(
    **config,
) -> Generator['TableIndexAnalyzer', None, None]:
    with create_browser(**config) as browser:
        yield TableIndexAnalyzer(browser.engine)



from .abstract import AbstractBrowser
from contextlib import contextmanager
from recap.plugins import load_browser_plugins
from typing import Generator


@contextmanager
def create_browser(
    plugin: str,
    **config,
) -> Generator['AbstractBrowser', None, None]:
    browser_plugins = load_browser_plugins()
    if browser_module := browser_plugins.get(plugin):
        with browser_module.create_browser(**config) as browser:
            yield browser



import logging
import sqlalchemy
from contextlib import contextmanager
from recap.analyzers.abstract import AbstractAnalyzer, BaseMetadataModel
from recap.browsers.db import create_browser, TablePath, ViewPath
from typing import Generator


log = logging.getLogger(__name__)


class Comment(BaseMetadataModel):
    __root__: str | None = None


class TableCommentAnalyzer(AbstractAnalyzer):
    def __init__(self, engine: sqlalchemy.engine.Engine):
        self.engine = engine

    def analyze(
        self,
        path: TablePath | ViewPath,
    ) -> Comment | None:
        table = path.table if isinstance(path, TablePath) else path.view
        comment = sqlalchemy.inspect(self.engine).get_table_comment(
            table,
            path.schema_,
        )
        comment_text = comment.get('text')
        if comment_text:
            return Comment.parse_obj(comment_text)
        return None


@contextmanager
def create_analyzer(**config) -> Generator['TableCommentAnalyzer', None, None]:
    with create_browser(**config) as browser:
        yield TableCommentAnalyzer(browser.engine)



import logging
import typer
import sys
from fastapi import APIRouter
from types import ModuleType


log = logging.getLogger(__name__)


if sys.version_info < (3, 10):
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points


ANALYZER_PLUGIN_GROUP = 'recap.analyzers'
BROWSER_PLUGIN_GROUP = 'recap.browsers'
CATALOG_PLUGIN_GROUP = 'recap.catalogs'
COMMAND_PLUGIN_GROUP = 'recap.commands'
ROUTER_PLUGIN_GROUP = 'recap.routers'


def load_analyzer_plugins() -> dict[str, ModuleType]:
    plugins = {}
    analyzer_plugins = entry_points(group=ANALYZER_PLUGIN_GROUP)
    for analyzer_plugin in analyzer_plugins:
        analyzer_plugin_name = analyzer_plugin.name
        try:
            analyzer_plugin_class = analyzer_plugin.load()
            plugins[analyzer_plugin_name] = analyzer_plugin_class
        except ImportError as e:
            log.debug(
                "Skipping analyzer=%s due to import error.",
                analyzer_plugin_name,
                exc_info=e,
            )
    return plugins


def load_browser_plugins() -> dict[str, ModuleType]:
    plugins = {}
    browser_plugins = entry_points(group=BROWSER_PLUGIN_GROUP)
    for browser_plugin in browser_plugins:
        browser_plugin_name = browser_plugin.name
        try:
            browser_plugin_class = browser_plugin.load()
            plugins[browser_plugin_name] = browser_plugin_class
        except ImportError as e:
            log.debug(
                "Skipping browser=%s due to import error.",
                browser_plugin_name,
                exc_info=e,
            )
    return plugins


def load_catalog_plugins() -> dict[str, ModuleType]:
    plugins = {}
    catalog_plugins = entry_points(group=CATALOG_PLUGIN_GROUP)
    for catalog_plugin in catalog_plugins:
        catalog_plugin_name = catalog_plugin.name
        try:
            catalog_plugin_class = catalog_plugin.load()
            plugins[catalog_plugin_name] = catalog_plugin_class
        except ImportError as e:
            log.debug(
                "Skipping catalog=%s due to import error.",
                catalog_plugin_name,
                exc_info=e,
            )
    return plugins


def load_command_plugins() -> dict[str, typer.Typer]:
    plugins = {}
    command_plugins = entry_points(group=COMMAND_PLUGIN_GROUP)
    for command_plugin in command_plugins:
        command_plugin_name = command_plugin.name
        try:
            command_plugin_instance = command_plugin.load()
            plugins[command_plugin_name] = command_plugin_instance
        except ImportError as e:
            log.debug(
                "Skipping command=%s due to import error.",
                command_plugin_name,
                exc_info=e,
            )
    return plugins


def load_router_plugins() -> dict[str, APIRouter]:
    plugins = {}
    router_plugins = entry_points(group=ROUTER_PLUGIN_GROUP)
    for router_plugin in router_plugins:
        router_plugin_name = router_plugin.name
        try:
            router_plugin_instance = router_plugin.load()
            plugins[router_plugin_name] = router_plugin_instance
        except ImportError as e:
            log.debug(
                "Skipping router=%s due to import error.",
                router_plugin_name,
                exc_info=e,
            )
    return plugins


def init_command_plugins(app: typer.Typer) -> typer.Typer:
    plugins = load_command_plugins()

    for command_plugin_name, command_plugin in plugins.items():
        # If the plugin has a single command, then put it directly into the
        # current Typer app. If the plugin has multiple commands, then
        # treat it as a command group, and add it as such.
        # TODO Shouldn't need to do this, but Typer has a bug.
        # https://github.com/tiangolo/typer/issues/119
        if len(command_plugin.registered_commands) == 1:
            callback = command_plugin.registered_commands[0].callback
            app.command(command_plugin_name)(callback) # pyright: ignore [reportGeneralTypeIssues]
        else:
            app.add_typer(command_plugin, name=command_plugin_name)

    return app



import inspect
from datetime import datetime
from fastapi import APIRouter, Body, Depends, HTTPException
from pydantic import BaseModel
from recap.catalogs.abstract import AbstractCatalog
from recap.paths import CatalogPath
from recap.plugins import load_browser_plugins
from recap.server import get_catalog
from recap.typing import (
    BrowserInspector,
    ModuleInspector,
    get_pydantic_model_for_path,
)
from typing import Callable


router = APIRouter(
    prefix="/catalog"
)


def add_route(
    router: APIRouter,
    endpoint: Callable,
    method: str,
    metadata_class: type[BaseModel],
    browser_root_path: type[CatalogPath],
    child_path: type[CatalogPath],
):
    dynamic_params = []
    path_fields = browser_root_path.__fields__ | child_path.__fields__
    endpoint_signature = inspect.signature(endpoint)
    endpoint_existing_params = [
        p for p in endpoint_signature.parameters.values()
        if p.kind != inspect.Parameter.VAR_KEYWORD
    ]

    for field in path_fields.values():
        dynamic_params.append(
            inspect.Parameter(
                field.alias or field.name,
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                annotation=field.type_,
                default=field.default,
            )
        )

    endpoint.__signature__ = inspect.Signature(
        dynamic_params + endpoint_existing_params
    )

    router.add_api_route(
        path=browser_root_path.template + child_path.template + '/metadata',
        endpoint=endpoint,
        dependencies=[Depends(get_catalog)],
        response_model=metadata_class if method == 'GET' else None,
        methods=[method],
    )


def add_routes(
    router: APIRouter,
    metadata_class: type[BaseModel],
    browser_root_path: type[CatalogPath],
    child_path: type[CatalogPath],
):
    def read_metadata(
        as_of: datetime | None = None,
        catalog: AbstractCatalog = Depends(get_catalog),
        **kwargs,
    ) -> metadata_class:
        path = (browser_root_path.template + child_path.template).format(**kwargs)
        metadata = catalog.read(path, as_of)
        if metadata:
            return metadata_class.parse_obj(metadata)
        raise HTTPException(status_code=404)

    def put_metadata(
        metadata: metadata_class = Body(),
        catalog: AbstractCatalog = Depends(get_catalog),
        **kwargs,
    ):
        path = (browser_root_path.template + child_path.template).format(**kwargs)
        metadata_dict = metadata.dict(
            by_alias=True,
            exclude_defaults=True,
            exclude_none=True,
            exclude_unset=True,
        )
        catalog.write(path, metadata_dict)

    def patch_metadata(
        metadata: metadata_class = Body(),
        catalog: AbstractCatalog = Depends(get_catalog),
        **kwargs,
    ):
        path = (browser_root_path.template + child_path.template).format(**kwargs)
        metadata_dict = metadata.dict(
            by_alias=True,
            exclude_defaults=True,
            exclude_none=True,
            exclude_unset=True,
        )
        catalog.write(path, metadata_dict, True)

    add_route(
        router,
        read_metadata,
        'GET',
        metadata_class,
        browser_root_path,
        child_path,
    )

    add_route(
        router,
        put_metadata,
        'PUT',
        metadata_class,
        browser_root_path,
        child_path,
    )

    add_route(
        router,
        patch_metadata,
        'PATCH',
        metadata_class,
        browser_root_path,
        child_path,
    )


def add_metadata_paths(router: APIRouter):
    for browser_module in load_browser_plugins().values():
        module_inspector = ModuleInspector(browser_module)
        if browser_class := module_inspector.browser_type():
            browser_inspector = BrowserInspector(browser_class)
            root_path_type = browser_inspector.root_type()
            for child_path_class in browser_inspector.children_types():
                metadata_class = get_pydantic_model_for_path(child_path_class)
                if root_path_type and metadata_class:
                    add_routes(
                        router,
                        metadata_class,
                        root_path_type,
                        child_path_class,
                    )


add_metadata_paths(router)



import logging
import sqlalchemy
from contextlib import contextmanager
from recap.analyzers.abstract import AbstractAnalyzer, BaseMetadataModel
from recap.browsers.db import create_browser, TablePath, ViewPath
from typing import Generator


log = logging.getLogger(__name__)


class ForeignKey(BaseMetadataModel):
    constrained_columns: list[str]
    referred_columns: list[str]
    referred_schema: str
    referred_table: str


class ForeignKeys(BaseMetadataModel):
    __root__: dict[str, ForeignKey] = {}


class TableForeignKeyAnalyzer(AbstractAnalyzer):
    def __init__(self, engine: sqlalchemy.engine.Engine):
        self.engine = engine

    def analyze(
        self,
        path: TablePath | ViewPath,
    ) -> ForeignKeys | None:
        table = path.table if isinstance(path, TablePath) else path.view
        results = {}
        fks = sqlalchemy.inspect(self.engine).get_foreign_keys(
            table,
            path.schema_,
        )
        for fk_dict in fks or []:
            results[fk_dict['name']] = ForeignKey(
                constrained_columns=fk_dict['constrained_columns'],
                referred_columns=fk_dict['referred_columns'],
                referred_schema=fk_dict['referred_schema'],
                referred_table=fk_dict['referred_table'],
            )
        if results:
            return ForeignKeys.parse_obj(results)
        return None


@contextmanager
def create_analyzer(
    **config,
) -> Generator['TableForeignKeyAnalyzer', None, None]:
    with create_browser(**config) as browser:
        yield TableForeignKeyAnalyzer(browser.engine)



import logging
import sqlalchemy
from contextlib import contextmanager
from recap.analyzers.abstract import AbstractAnalyzer, BaseMetadataModel
from recap.browsers.db import create_browser, ViewPath
from typing import Generator


log = logging.getLogger(__name__)


class ViewDefinition(BaseMetadataModel):
    __root__: str | None = None


class TableViewDefinitionAnalyzer(AbstractAnalyzer):
    def __init__(self, engine: sqlalchemy.engine.Engine):
        self.engine = engine

    def analyze(
        self,
        path: ViewPath,
    ) -> ViewDefinition | None:
        # TODO sqlalchemy-bigquery doesn't work right with this API
        # https://github.com/googleapis/python-bigquery-sqlalchemy/issues/539
        view = path.view
        if self.engine.dialect.name == 'bigquery':
            view = f"{path.schema_}.{path.view}"
        def_dict = sqlalchemy \
            .inspect(self.engine) \
            .get_view_definition(view, path.schema_)
        if def_dict:
            return ViewDefinition.parse_obj(def_dict)
        return None


@contextmanager
def create_analyzer(
    **config,
) -> Generator['TableViewDefinitionAnalyzer', None, None]:
    with create_browser(**config) as browser:
        yield TableViewDefinitionAnalyzer(browser.engine)



import typer
from recap.config import settings
from recap.logging import setup_logging
from recap.server import fastapp


app = typer.Typer()


@app.command()
def serve():
    """
    Starts a FastAPI server that exposes a catalog API over HTTP/JSON.
    """

    import uvicorn

    uvicorn.run(
        fastapp,
        log_config=setup_logging(),
        **settings('api', {}),
    )



import typer
from recap import plugins
from rich import print_json


app = typer.Typer(help="Lists available plugins.")


@app.command()
def analyzers():
    """
    Lists all analyzer plugins.
    """

    print_json(data=list(plugins.load_analyzer_plugins().keys()))


@app.command()
def browsers():
    """
    Lists all browser plugins.
    """

    print_json(data=list(plugins.load_browser_plugins().keys()))


@app.command()
def catalogs():
    """
    Lists all catalog plugins.
    """

    print_json(data=list(plugins.load_catalog_plugins().keys()))


@app.command()
def commands():
    """
    Lists all command plugins.
    """

    print_json(data=list(plugins.load_command_plugins().keys()))


@app.command()
def routers():
    """
    Lists all router plugins.
    """

    print_json(data=list(plugins.load_router_plugins().keys()))



from pydantic import BaseModel, create_model
from recap.analyzers.abstract import AbstractAnalyzer, BaseMetadataModel
from recap.browsers import AbstractBrowser
from recap.paths import CatalogPath
from recap.plugins import load_analyzer_plugins
from types import ModuleType
from typing import get_type_hints


class AnalyzerInspector:
    def __init__(self, analyzer_class: type[AbstractAnalyzer]):
        self.analyzer_class = analyzer_class

    def input_path_types(self) -> list[type[CatalogPath]]:
        input_paths = []
        analyze_method = self.analyzer_class.analyze
        if analyze_paths := get_type_hints(analyze_method).get('path'):
            type_stack = [analyze_paths]
            while type_stack:
                child_type = type_stack.pop()
                try: # EAFP
                    # Not all types are classes, so issubclass will barf sometimes
                    if issubclass(child_type, CatalogPath):
                        input_paths.append(child_type)
                except:
                    pass
                if hasattr(child_type, '__args__'):
                    type_stack.extend(child_type.__args__) # pyright: ignore [reportGeneralTypeIssues]
        return input_paths

    def return_type(self) -> type[BaseMetadataModel] | None:
        analyze_method = self.analyzer_class.analyze
        if return_model := get_type_hints(analyze_method).get('return'):
            type_stack = [return_model]
            while type_stack:
                child_type = type_stack.pop()
                try: # EAFP
                    # Not all types are classes, so issubclass will barf sometimes
                    if issubclass(child_type, BaseMetadataModel):
                        return child_type
                except:
                    pass
                if hasattr(child_type, '__args__'):
                    type_stack.extend(child_type.__args__) # pyright: ignore [reportGeneralTypeIssues]
        return None


class BrowserInspector:
    def __init__(self, browser_class: type[AbstractBrowser]):
        self.browser_class = browser_class

    def children_types(self) -> list[type[CatalogPath]]:
        return_paths = []
        children_method = self.browser_class.children
        children_type_hints = get_type_hints(children_method)
        if children_returns := children_type_hints.get('return'):
            type_stack = [children_returns]
            while type_stack:
                child_type = type_stack.pop()
                try: # EAFP
                    # Not all types are classes, so issubclass will barf sometimes
                    if issubclass(child_type, CatalogPath):
                        return_paths.append(child_type)
                except:
                    pass
                if hasattr(child_type, '__args__'):
                    type_stack.extend(child_type.__args__) # pyright: ignore [reportGeneralTypeIssues]
        return return_paths

    def root_type(self) -> type[CatalogPath] | None:
        root_method = self.browser_class.root
        if return_model := get_type_hints(root_method).get('return'):
            type_stack = [return_model]
            while type_stack:
                child_type = type_stack.pop()
                try: # EAFP
                    # Not all types are classes, so issubclass will barf sometimes
                    if issubclass(child_type, CatalogPath):
                        return child_type
                except:
                    pass
                if hasattr(child_type, '__args__'):
                    type_stack.extend(child_type.__args__) # pyright: ignore [reportGeneralTypeIssues]
        return None


class ModuleInspector:
    def __init__(self, module: ModuleType):
        self.module = module

    def browser_type(self) -> type[AbstractBrowser] | None:
        browser_module_return = get_type_hints(
            self.module.create_browser,
        ).get('return')
        if browser_module_return:
            # browser_module_return is Generator[SomeBrowserType, None, None].
            # __args__[0] gets SomeBrowserType from the Generator type.
            return browser_module_return.__args__[0]
        return None

    def analyzer_type(self) -> type[AbstractAnalyzer] | None:
        analyzer_module_return = get_type_hints(
            self.module.create_analyzer,
        ).get('return')
        if analyzer_module_return:
            # analyzer_module_return is Generator[SomeAnalyzerType, None, None].
            # __args__[0] gets SomeAnalyzerType from the Generator type.
            return analyzer_module_return.__args__[0]
        return None


def get_analyzers_for_path(path_class: type[CatalogPath]) -> list[type[AbstractAnalyzer]]:
    analyzers = []
    for analyzer_module in load_analyzer_plugins().values():
        module_inspector = ModuleInspector(analyzer_module)
        if analyzer_class := module_inspector.analyzer_type():
            analyzer_inspector = AnalyzerInspector(analyzer_class)
            analyzer_input_paths = analyzer_inspector.input_path_types()
            if path_class in analyzer_input_paths:
                analyzers.append(analyzer_class)
    return analyzers


def get_pydantic_model_for_path(
    path_class: type[CatalogPath],
    name: str | None = None,
) -> type[BaseModel] | None:
    path_attrs = {}
    for analyzer_class in get_analyzers_for_path(path_class):
        analyzer_inspector = AnalyzerInspector(analyzer_class)
        if analyzer_return_model := analyzer_inspector.return_type():
            path_attrs[analyzer_return_model.key()] = (analyzer_return_model, None)
    return create_model(
        name or f"{path_class.__name__}Metadata",
        **path_attrs,
    ) if path_attrs else None



import typer
from .logging import setup_logging
from .plugins import init_command_plugins


LOGGING_CONFIG = setup_logging()
app = init_command_plugins(typer.Typer())


if __name__ == "__main__":
    app()



from abc import ABC, abstractmethod
from recap.paths import CatalogPath


class AbstractBrowser(ABC):
    """
    The abstract class for all browsers. Recap uses a browser abstraction to
    deal with different data infrastructure in a standard way.

    Browsers map infrastructure objects into a directory format. A different
    browser is used for each type of infrastructure.

    A DatabaseBrowser might list directories like this:

        /schemas
        /schemas/public
        /schemas/public/tables
        /schemas/public/tables/groups
        /schemas/public/tables/users
    """

    @abstractmethod
    def children(self, path: str) -> list[CatalogPath] | None:
        """
        Given a path, returns its children. Using the example above,
        path=/schemas/public/tables would return:

            [
                TablePath(
                    schema='public',
                    table='groups',
                ),
                TablePath(
                    schema='public',
                    table='users',
                ),
            ]

        The path parameter is relative; it does not include the browser's root.
        """

        raise NotImplementedError

    @abstractmethod
    def root(self) -> CatalogPath:
        """
        The root path for this browser. The root path is prefixed to all
        paths when persisting metdata to a catalog.

        Using the example above, root might return something like:

            /databases/postgresql/instances/prd-db

        Thus, the full catalog path for the `users` table would be:

            /databases/postgresql/instances/prd-db/schemas/public/tables/users
        """
        raise NotImplementedError



import logging
import sqlalchemy
from contextlib import contextmanager
from recap.analyzers.abstract import AbstractAnalyzer, BaseMetadataModel
from recap.browsers.db import create_browser, TablePath, ViewPath
from typing import Generator


log = logging.getLogger(__name__)


class PrimaryKey(BaseMetadataModel):
    name: str
    constrained_columns: list[str]


class TablePrimaryKeyAnalyzer(AbstractAnalyzer):
    def __init__(self, engine: sqlalchemy.engine.Engine):
        self.engine = engine

    def analyze(
        self,
        path: TablePath | ViewPath,
    ) -> PrimaryKey | None:
        table = path.table if isinstance(path, TablePath) else path.view
        pk_dict = sqlalchemy.inspect(self.engine).get_pk_constraint(
            table,
            path.schema_,
        )
        if pk_dict and pk_dict.get('name'):
            return PrimaryKey.parse_obj(pk_dict)
        return None


@contextmanager
def create_analyzer(
    **config,
) -> Generator['TablePrimaryKeyAnalyzer', None, None]:
    with create_browser(**config) as browser:
        yield TablePrimaryKeyAnalyzer(browser.engine)



import httpx
from .abstract import AbstractCatalog
from contextlib import contextmanager
from datetime import datetime
from recap.server import DEFAULT_URL
from typing import Any, Generator



class RecapCatalog(AbstractCatalog):
    """
    RecapCatalog makes HTTP requests to a Recap server and returns the results.
    """

    def __init__(
        self,
        client: httpx.Client,
    ):
        self.client = client

    def touch(
        self,
        path: str,
    ):
        self.write(path, {})

    def write(
        self,
        path: str,
        metadata: dict[str, Any],
        patch: bool = True,
    ):
        method = self.client.patch if patch else self.client.put
        response = method(
            f"/catalog{path}/metadata",
            json=metadata,
        )
        response.raise_for_status()

    def rm(
        self,
        path: str,
    ):
        self.client.delete(f"/catalog{path}").raise_for_status()

    def ls(
        self,
        path: str,
        as_of: datetime | None = None,
    ) -> list[str] | None:
        params: dict[str, Any] = {}
        if as_of:
            params['as_of'] = as_of.isoformat()
        response = self.client.get(f"/catalog{path}/children", params=params)
        if response.status_code == httpx.codes.OK:
            return response.json()
        if response.status_code == httpx.codes.NOT_FOUND:
            return None
        response.raise_for_status()

    def read(
        self,
        path: str,
        as_of: datetime | None = None,
    ) -> dict[str, Any] | None:
        params: dict[str, Any] = {}
        if as_of:
            params['as_of'] = as_of.isoformat()
        response = self.client.get(f"/catalog{path}/metadata", params=params)
        if response.status_code == httpx.codes.OK:
            return response.json()
        if response.status_code == httpx.codes.NOT_FOUND:
            return None
        response.raise_for_status()

    def search(
        self,
        query: str,
        as_of: datetime | None = None,
    ) -> list[dict[str, Any]]:
        params: dict[str, Any] = {'query': query}
        if as_of:
            params['as_of'] = as_of.isoformat()
        return self.client.get('/catalog', params=params).json()


@contextmanager
def create_catalog(
    url: str | None = None,
    **_,
) -> Generator['RecapCatalog', None, None]:
    with httpx.Client(base_url=url or DEFAULT_URL) as client:
        yield RecapCatalog(client)



import fnmatch
import logging
from .plugins import load_analyzer_plugins, load_browser_plugins
from contextlib import contextmanager, ExitStack
from pathlib import PurePosixPath
from recap.analyzers import create_analyzer
from recap.analyzers.abstract import AbstractAnalyzer
from recap.browsers import create_browser
from recap.browsers.abstract import AbstractBrowser
from recap.catalogs.abstract import AbstractCatalog
from recap.paths import CatalogPath, RootPath
from typing import Any, Generator


log = logging.getLogger(__name__)


class Crawler:
    """
    Crawler does three things:

    1. Browses the configured infrastructure
    2. Analyzes the infrastructure's data to generate metadata
    3. Stores the metadata in a catalog

    Recap's crawler is very simple right now. The crawler recursively browses
    and analyzes all children starting from an infrastructre's root location.
    """

    def __init__(
        self,
        browser: AbstractBrowser,
        catalog: AbstractCatalog,
        analyzers: list[AbstractAnalyzer],
        filters: list[str] = [],
    ):
        """
        :param browser: The browser to use when listing a path's children.
        :param catalog: The catalog to create directories and store metadata.
        :param analyzers: Analyzers to inspect each path for metadata.
        :param filters: Path filter to include only certain paths. Recap uses
            Unix filename pattern matching as defined in Python's fnmatch
            module. The path that's filtered doesn't include the root.
        """

        self.browser = browser
        self.catalog = catalog
        self.analyzers = analyzers
        self.filters = filters
        self.exploded_filters = self._explode_filters(filters)

    def crawl(self):
        log.info('Beginning crawl root=%s', self.browser.root())
        path_stack: list[CatalogPath] = [RootPath()]

        while len(path_stack) > 0:
            relative_path = path_stack.pop()
            full_path_posix = PurePosixPath(
                str(self.browser.root()),
                str(relative_path)[1:],
            )

            log.info("Crawling path=%s", relative_path)

            # 1. Read and save metadata for path if filters match.
            if self._matches(relative_path, self.filters):
                metadata = self._get_metadata(relative_path)
                self._write_metadata(full_path_posix, metadata)

            # 2. Add children (that match filter) to path_stack.
            children = self.browser.children(str(relative_path)) or []
            filtered_children = filter(
                lambda p: self._matches(p, self.exploded_filters),
                children,
            )
            path_stack.extend(filtered_children)

            # 3. Remove deleted children from catalog.
            self._remove_deleted(full_path_posix, children)

        log.info('Finished crawl root=%s', self.browser.root())

    def _matches(
        self,
        relative_path: CatalogPath,
        filters: list[str],
    ) -> bool:
        """
        Check if a path matches any filters.

        :returns: True if path matches a filter or if filters is empty.
        """

        for filter in filters:
            if fnmatch.fnmatch(str(relative_path), filter):
                return True
        return False if filters else True

    def _get_metadata(
        self,
        path: CatalogPath,
    ) -> dict[str, Any]:
        """
        Run all analyzers on a path.

        :returns: A dictionary of all metadata returned from all analyzers.
        """

        results = {}
        for analyzer in self.analyzers:
            log.debug(
                'Analyzing path=%s analyzer=%s',
                path,
                analyzer.__class__.__name__,
            )
            try: # EAFP
                if metadata := analyzer.analyze(path):
                    metadata_dict = metadata.dict(
                        by_alias=True,
                        exclude_none=True,
                        exclude_unset=True,
                        exclude_defaults=True,
                    )
                    # Have to unpack __root__ if it exists, sigh.
                    # https://github.com/pydantic/pydantic/issues/1193
                    metadata_dict = metadata_dict.get(
                        '__root__',
                        metadata_dict
                    )
                    results |= {metadata.key(): metadata_dict}
            except Exception as e:
                log.debug(
                    'Unable to process path with analyzer path=%s analyzer=%s',
                    path,
                    analyzer.__class__.__name__,
                    exc_info=e,
                )
        return results

    def _write_metadata(
        self,
        full_path_posix: PurePosixPath,
        metadata: dict[str, Any],
    ):
        """
        Write a metadata dictionary to a path in the catalog.
        """

        log.debug(
            'Writing metadata path=%s metadata=%s',
            full_path_posix,
            metadata,
        )
        self.catalog.write(str(full_path_posix), metadata, True)

    def _remove_deleted(
        self,
        full_path_posix: PurePosixPath,
        instance_children: list[CatalogPath],
    ):
        """
        Compares the path's children in the browser vs. what is currently in
        the catalog. Deletes all children that appear in the catalog, but no
        longer appear in the browser. This behavior removes children that used
        to exist in data infrastructure, but have been deleted since the last
        crawl.
        """

        catalog_children = self.catalog.ls(str(full_path_posix)) or []
        instance_children_names = [c.name() for c in instance_children]
        # Find catalog children that are not in the browser's children.
        deleted_children = [
            catalog_child
            for catalog_child in catalog_children
            if catalog_child not in instance_children_names
        ]
        for child in deleted_children:
            path_to_remove = str(PurePosixPath(full_path_posix, child))
            log.debug('Removing deleted path from catalog: %s', path_to_remove)
            self.catalog.rm(path_to_remove)

    def _explode_filters(self, filters: list[str]) -> list[str]:
        """
        Returns a list of paths that bread-crumb from the filter all the way
        back to root. For example:

            filters=[
                '/**/schemas/my_db/tables/foo*'
            ]
            returns=[
                '/**',
                '/**/schemas',
                '/**/schemas/my_db',
                '/**/schemas/my_db/tables',
                '/**/schemas/my_db/tables/foo*',
            ]

        We need to do this so that parents match the filter and crawling
        reaches the wild-carded children.
        """

        exploded_filters = []
        for filter in filters:
            fragments = filter.split('/')
            partial_path = PurePosixPath('/')
            for fragment in fragments:
                partial_path = PurePosixPath(partial_path, fragment)
                exploded_filters.extend(str(partial_path))
        return exploded_filters

@contextmanager
def create_crawler(
    catalog: AbstractCatalog,
    **config,
) -> Generator['Crawler', None, None]:
    analyzer_plugins = load_analyzer_plugins()
    browser_plugins = load_browser_plugins()
    url = config.get('url')
    excludes = config.get('excludes', [])
    filters = config.get('filters', [])

    assert url, \
        f"No url defined for instance config={config}"

    analyzers = []
    browser = None

    with ExitStack() as stack:
        for browser_name in browser_plugins.keys():
            try:
                browser_context_manager = create_browser(
                    plugin=browser_name,
                    **config,
                )
                browser = stack.enter_context(browser_context_manager)

                # If we got this far, we found a browser. Stop looking.
                break
            except Exception as e:
                    log.debug(
                        'Skipped browser for url=%s name=%s',
                        url,
                        browser_name,
                        exc_info=e,
                    )

        assert browser, f"Found no browser for url={url}"

        for analyzer_name in analyzer_plugins.keys():
            if (analyzer_name not in excludes):
                try:
                    analyzer_context_manager = create_analyzer(
                        plugin=analyzer_name,
                        **config,
                    )
                    analyzer = stack.enter_context(analyzer_context_manager)
                    analyzers.append(analyzer)
                except Exception as e:
                    log.debug(
                        'Skipped analyzer for url=%s name=%s',
                        url,
                        analyzer_name,
                        exc_info=e,
                    )

        assert analyzers, f"Found no analyzers for url={url}"

        yield Crawler(
            browser,
            catalog,
            analyzers,
            filters,
        )
